import numpy as np
import os
import sys
sys.path.insert(0, 'src')

import scinumtools as snt

def test_data_combinations():
    
    pc = snt.DataCombination([
        ['a','b'],
        ['c','d','e']
    ])

    # test keys
    keys = [(0, 0),
            (0, 1),
            (0, 2),
            (1, 0),
            (1, 1),
            (1, 2)]
    for i,key in enumerate(pc.keys()):
        assert keys[i] == key

    # test values
    values = [('a', 'c'),
              ('a', 'd'),
              ('a', 'e'),
              ('b', 'c'),
              ('b', 'd'),
              ('b', 'e')]
    for i,value in enumerate(pc.values()):
        assert values[i] == value

    # test items
    for i,(key,value) in enumerate(pc.items()):
        assert keys[i] == key
        assert values[i] == value



import sys, os
import pytest
import numpy as np
sys.path.insert(0, 'src')

from scinumtools.dip import DIP
from scinumtools.dip.settings import Format
from scinumtools.dip.datatypes import FloatType, IntegerType, StringType, BooleanType

def parse(code):
    with DIP() as p:
        p.from_string(code)
        return p.parse().data(verbose=True,format=Format.TYPE)

def test_name():
    data = parse('''
very-long.node23_NAME int = 1
    ''')
    np.testing.assert_equal(data,{
        'very-long.node23_NAME': IntegerType(1),
    })
    with pytest.raises(Exception) as e_info:
        parse('wrong$name int = 3')
    assert e_info.value.args[0] == "Name has an invalid format: wrong$name int = 3"
    
def test_types():
    data = parse('''
adult bool = true
age int = 20 yr
weight float = 63.3 kg
name str = 'Laura'
    ''')
    np.testing.assert_equal(data,{
        'adult':  BooleanType(True),
        'age':    IntegerType(20, 'yr'),
        'weight': FloatType(63.3, 'kg'),
        'name':   StringType('Laura')
    })
    with pytest.raises(Exception) as e_info:
        parse('age bool = true a')
    assert e_info.value.args[0] == "Boolean datatype does not support units:"
    assert e_info.value.args[1] == "age bool = true a"
    with pytest.raises(Exception) as e_info:
        parse("age str = 23 a")
    assert e_info.value.args[0] == "String datatype does not support units:"
    assert e_info.value.args[1] == "age str = 23 a"

def test_dimensions():
    data = parse('''
counts int[3] = [4234,34,2]
lengths float[2:,2] = [[4234,34],[234,34]] cm
colleagues str[:] = ["John","Patricia","Lena"]
logic bool[2] = [true,false]
    ''')
    np.testing.assert_equal(data,{
        'counts':      IntegerType([4234,   34,    2]),
        'lengths':     FloatType([[4234.,   34.],[ 234.,   34.]], 'cm'),
        'colleagues':  StringType(['John', 'Patricia', 'Lena']),
        'logic':       BooleanType([ True, False])
    })
    with pytest.raises(Exception) as e_info:
        parse('counts int[2] = [4234,34,2]')
    assert e_info.value.args[0] == "Node 'counts' has invalid dimension: dim(0)=3 > 2"
    with pytest.raises(Exception) as e_info:
        parse('counts int[2] = [4234]')
    assert e_info.value.args[0] == "Node 'counts' has invalid dimension: dim(0)=1 < 2"
    with pytest.raises(Exception) as e_info:
        parse('counts int[:2] = [4234,34,2]')
    assert e_info.value.args[0] == "Node 'counts' has invalid dimension: dim(0)=3 > 2"
    with pytest.raises(Exception) as e_info:
        parse('counts int[2:] = [4234]')
    assert e_info.value.args[0] == "Node 'counts' has invalid dimension: dim(0)=1 < 2"
    with pytest.raises(Exception) as e_info:
        parse('counts int[2,3:] = [[234,4234],[234,34]]')
    assert e_info.value.args[0] == "Node 'counts' has invalid dimension: dim(1)=2 < 3"
    with pytest.raises(Exception) as e_info:
        parse('counts int = [[234,4234],[234,34]]')
    assert e_info.value.args[0] == "Could not convert raw value to type:"
def test_declarations():
    data = parse('''
cash bool
cash = true
weight float kg
weight = 77
''')
    np.testing.assert_equal(data,{
        'cash':   BooleanType(True),
        'weight': FloatType(77, 'kg')
    })
    with pytest.raises(Exception) as e_info:
        parse('counts int')
    assert e_info.value.args[0] == "Node value must be defined:"

def test_strings():
    data = parse('''
country str = Canada              # strings without whitespace
name str = "Johannes Brahms"      # strings with a whitespace
counts int[3] = "[0, 1, 2]"       # arrays with whitespaces between items
answers bool[2] = "[true, false]" 
names str[2] = '["Jolana", "Anastasia"]'
girl_friend str = "\"l'amie\""    # escaping of double quotes
boy_friend str = '"l\'ami"'       # escaping of single quotes
hashtag str = '#nocomment'        # comment
anticommutator str = '{a,b}'      # this is not an import
    ''')
    np.testing.assert_equal(data,{
        'country':        StringType('Canada'),
        'name':           StringType('Johannes Brahms'),
        'counts':         IntegerType([0, 1, 2]),
        'answers':        BooleanType([ True, False]),
        'names':          StringType(['Jolana', 'Anastasia']),
        'girl_friend':    StringType('"l\'amie"'),
        'boy_friend':     StringType('"l\'ami"'),
        'hashtag':        StringType('#nocomment'),
        'anticommutator': StringType('{a,b}')
      })
    with pytest.raises(Exception) as e_info:
        parse('name str = Johannes Brahms')
    assert e_info.value.args[0] == "String datatype does not support units:"
    assert e_info.value.args[1] == "name str = Johannes Brahms"
        
if __name__ == "__main__":
    # Specify wich test to run
    test = sys.argv[1] if len(sys.argv)>1 else True

    # Loop through all tests
    for fn in dir(sys.modules[__name__]):
        if fn[:5]=='test_' and (test is True or test==fn[5:]):
            print(f"\nTesting: {fn}\n")
            locals()[fn]()



import sys, os
import pytest
import numpy as np
sys.path.insert(0, 'src')

from scinumtools.dip import DIP
from scinumtools.dip.solvers import NumericalSolver

def test_plus_minus():
    with NumericalSolver() as p:
        assert p.equal('2 + 4 - 3',                '3')           # numbers only
        assert p.equal('1 - -3 + -4',              '0')           # negative values and zeros
        assert p.equal('34 cm + 4 mm',             '34.4 cm')     # addition
        assert p.equal('10 m + 4 cm + 3 m + 1 mm', '13.041 m')    # multiple additions
        assert p.equal('3 m - 5 cm',               '295 cm')      # substraction
        assert p.equal('10 m - 1 m + 3 cm - 3 mm', '9.027 m')     # multiple substractions
        assert p.equal('23 kg*m*s-2',              '23 kg*m*s-2') # complex units with -
    with pytest.raises(Exception) as e_info:
        p.solve('10 m + 1 J')
    assert e_info.value.args[0] == "Unsupported conversion between units:"
    with pytest.raises(Exception) as e_info:
        p.solve('10 m - 1 J')
    assert e_info.value.args[0] == "Unsupported conversion between units:"

def test_times_divide():
    with NumericalSolver() as p:
        assert p.equal('8 / 4 * 3',                    '6')          # numbers only
        assert p.equal('2 * 3 * 4',                    '24')         # multiple *
        assert p.equal('8 / 2 / 4',                    '1')          # multiple /
        assert p.equal('-8 / 2 * -4',                  '16')         # negative values
        assert p.equal('10 m * 2 cm',                  '0.2 m2')     # * with dimensional values
        assert p.equal('4 cm2 + 10 m * 2 cm - 0.2 m2', '0.0004 m2')  # * within + -
        assert p.equal('10 m2 / 200 cm',               '50 dm')      # / with dimensional values
        assert p.equal('4 m2 + 10 m3 / 2 m - 3 m2',    '6 m2')       # / within + -
        assert p.equal('3 kg * 4 m2 / 2 s2 + 1e7 erg', '7 J')        # combination of * /
        assert p.equal('23 kg*m2/s2 * 2',              '46 J')       # complex units with *
        assert p.equal('23 kg*m2/s2 / 2 J',            '11.5')       # complex units with /
        
def test_parenthesis():
    with NumericalSolver() as p:
        assert p.equal('(10 m - 1 m) + 3 cm - 3 mm',       '9.027 m')  # at the begining
        assert p.equal('10 m - (1 m + 3 cm) - 3 mm',       '8.967 m')  # in the middle
        assert p.equal('10 m - (1 m + 3 cm - 3 mm)',       '8.973 m')  # at the end
        assert p.equal('4 m2 + 10000 mm * (300 cm - 1 m)', '24 m2')    # after *
        assert p.equal('36 m2 / (20 dm * 300 cm) - 1',     '5')        # after \
        assert p.equal('(2 + (3 - 4))',                    '1')        # nested

def test_functions():
    with NumericalSolver() as p:
        # unary operations
        assert p.equal('exp(10 m / 5 cm)',    '7.22597376e86')   
        assert p.equal('log(10 m / 5 cm)',     '5.298317')   
        assert p.equal('log10(10 m / 5 cm)',  '2.301029995')   
        assert p.equal('sin(10 m / 5 cm)',    '-0.873297297')   
        assert p.equal('cos(10 m / 5 cm)',    '0.4871876750')   
        # binary operations
        assert p.equal('pow(10 m, 2)',        '100 m2')   

def test_references():
    with DIP() as dip:
        dip.from_string("""
a float = 10 m
b float = 300 cm
        """)
        env = dip.parse()
    with NumericalSolver(env) as p:
        assert p.equal('3 m * log10({?a} / (7 cm - 20 mm)) + {?b}', '9.9030899 m')
        
if __name__ == "__main__":
    # Specify wich test to run
    test = sys.argv[1] if len(sys.argv)>1 else True

    # Loop through all tests
    for fn in dir(sys.modules[__name__]):
        if fn[:5]=='test_' and (test is True or test==fn[5:]):
            print(f"\nTesting: {fn}\n")
            locals()[fn]()



import copy

from ..environment import Environment
from ..settings import Format
from ..datatypes import Type

class FunctionSolver:

    env: Environment

    def __init__(self, env:Environment=None, **kwargs):
        if env:
            self.env = env
        else:
            self.env = Environment()
    
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        pass

    def solve(self, fn_name:str, in_units=None):
        fn = self.env.functions[fn_name]
        data = copy.deepcopy(self.env.data(format=Format.TYPE))
        result = fn(data)
        if isinstance(result, Type):
            if in_units:
                result.convert(in_units)
            return result.value
        else:
            return result



from typing import Union, List
import numpy as np

from .type_number import NumberType

class IntegerType(NumberType):
    value: Union[int,list]
    dtype = int

    def __str__(self):
        if self.unit:
            return f"IntegerType({self.value} {self.unit})"
        else:
            return f"IntegerType({self.value})"
        
    def __init__(self, value, unit=None, **kwargs):
        if isinstance(value, IntegerType):
            kwargs['value'] = value.value
            kwargs['unit'] = value.unit
        else:
            kwargs['value'] = value
            kwargs['unit'] = unit
        if isinstance(kwargs['value'], np.ndarray):
            kwargs['value'] = kwargs['value'].tolist()
        super().__init__(**kwargs)



import sys, os
import pytest
import numpy as np
sys.path.insert(0, 'src')

from scinumtools.dip import DIP
from scinumtools.dip.settings import Format
from scinumtools.dip.datatypes import FloatType, IntegerType, StringType, BooleanType

def parse(code):
    with DIP() as p:
        p.from_string(code)
        return p.parse().data(verbose=True,format=Format.TYPE)

def test_invalid_start():
    with pytest.raises(Exception) as e_info:
        parse('''
        @end
        ''')
    assert e_info.value.args[0] == "Invalid condition:"
    assert e_info.value.args[1] == "@end"
    with pytest.raises(Exception) as e_info:
        parse('''
        @else
           car str = 'BMW'
        ''')
    assert e_info.value.args[0] == "Invalid condition:"
    assert e_info.value.args[1] == "@else"
    with pytest.raises(Exception) as e_info:
        parse('''
        @case true
          @end
        ''')
    assert e_info.value.args[0] == "Invalid condition:"
    assert e_info.value.args[1] == "@case.@end"
    
def test_nested_condition():
    data = parse('''
    @case false
      flower str = 'rose'
    @else
      flower str = 'dandelion'
      @case false
        color str = 'red'
      @case false
        color str = 'blue'
      @else
        @case true
          leaves int = 234
        color str = 'yellow'
    tree str = 'maple'
    ''')
    np.testing.assert_equal(data,{
        'flower': StringType('dandelion'),
        'leaves': IntegerType(234),
        'color':  StringType('yellow'),
        'tree':   StringType('maple'),
    })
    
def test_modifications():
    data = parse('''
    star str = 'Sun'

    @case false
      star = 'Sirius'
      nebula str = 'Orion'
    @else
      star = 'Wega'
      nebula str = 'Crab'

    nebula = 'Eagle'
    ''')
    np.testing.assert_equal(data,{
        'star':   StringType('Wega'),
        'nebula': StringType('Eagle'),
    })
    
def test_case():
    data = parse('''
    climate
      @case true                  # true condition
        warming bool = true       # nodes belonging to a case have higher indent
          increase float = 2 Cel  # subnodes also belong to the above case

      temperature float = 10.2 Cel   # case ends if indent of a new node is lower

    plant
      @case true                # first condition is true
        leaves int = 1302       
      @case false
        leaves int = 12304
      @end                      # case ends when explicitely terminated

    plant.@case false           # using compact node names
        flower str = 'green'
    plant.@case true            # second condition is true
        flower str = 'yellow'   
    plant.@else                 # else is not triggered
        flower str = 'red'

    animal
      @case false
        cat str = 'lion'
      @case false
        cat str = 'tiger'
      @else                     # return else, because none of the cases were true
        cat str = 'gepard'        
    ''')
    np.testing.assert_equal(data,{
        'climate.warming':          BooleanType(True),
        'climate.warming.increase': FloatType(2, 'Cel'),
        'climate.temperature':      FloatType(10.2, 'Cel'),
        'plant.leaves':             IntegerType(1302),
        'plant.flower':             StringType('yellow'),
        'animal.cat':               StringType('gepard'),
    })

def test_expression():
    data = parse('''
trafic

  # definitions
  limit float = 75 km/s 
  urban bool = true

  # first case with inline conditions
  @case ("{?trafic.limit} <= 50 km/s || {?trafic.urban}")

    road str = 'town'

  # second case with a block condition
  @case ("( {?trafic.limit} <= 100 km/s && {?trafic.limit} > 50 km/s )  && !{?trafic.urban}")

    road str = 'country'

  # if none of the cases apply
  @else

    road str = 'motorway'

  @end

  cars int = 12  # outside of case
    ''')
    np.testing.assert_equal(data,{
        'trafic.limit': FloatType(75, 'km/s'),
        'trafic.urban': BooleanType(True),
        'trafic.road':  StringType('town'),
        'trafic.cars':  IntegerType(12),
    })
  
if __name__ == "__main__":
    # Specify wich test to run
    test = sys.argv[1] if len(sys.argv)>1 else True

    # Loop through all tests
    for fn in dir(sys.modules[__name__]):
        if fn[:5]=='test_' and (test is True or test==fn[5:]):
            print(f"\nTesting: {fn}\n")
            locals()[fn]()



import sys, os
import pytest
import numpy as np
sys.path.insert(0, 'src')

from scinumtools.dip import DIP
from scinumtools.dip.settings import Format
from scinumtools.dip.datatypes import FloatType, IntegerType, StringType

def parse(code):
    with DIP() as p:
        p.from_string(code)
        return p.parse().data(verbose=True,format=Format.TYPE)
    
def test_inline_matrix():
    with DIP() as p:
        p.from_string('''
velocity int[1:,3] = """
[[42,34,35],
 [23,34,64],
 [35,23,23]]
""" km/s
        ''')
        env = p.parse()
    np.testing.assert_equal(env.data(verbose=True,format=Format.TYPE),{
        'velocity': IntegerType([[42,   34,   35],[ 23,   34,  64],[ 35, 23,  23]], 'km/s'),
    })

def test_inline_table():
    data = parse('''
outputs table = """
time float s
snapshot int
intensity float W/m2

0.234 0 2.34
1.355 1 9.4
2.535 2 3.4
3.255 3 2.3
4.455 4 23.4
  """  # endqotes can be indented
    ''')
    np.testing.assert_equal(data,{
        'outputs.time': FloatType([0.234, 1.355, 2.535, 3.255, 4.455], 's'),
        'outputs.snapshot': IntegerType([0, 1, 2, 3, 4]),
        'outputs.intensity': FloatType([ 2.34,  9.4 ,  3.4 ,  2.3 , 23.4 ], 'W/m2'),
    })

    
def test_inline_text():
    data = parse('''
text str = """
   tripple qotes # ' " \' \"
block of text
""" 
    ''')
    np.testing.assert_equal(data,{
        'text': StringType('   tripple qotes # \' " \' "\nblock of text')
    })
        
if __name__ == "__main__":
    # Specify wich test to run
    test = sys.argv[1] if len(sys.argv)>1 else True

    # Loop through all tests
    for fn in dir(sys.modules[__name__]):
        if fn[:5]=='test_' and (test is True or test==fn[5:]):
            print(f"\nTesting: {fn}\n")
            locals()[fn]()



import numpy as np
from typing import List, Dict, Union
from dataclasses import dataclass, field
import copy

from ..units import Quantity, UnitEnvironment
from .settings import Format, Sign, Keyword, Namespace
from .datatypes import NumberType

@dataclass
class EnvSource:
    source: Union[str] # can be also DIP
    path: str
    name: str
    
@dataclass
class Environment:
    # Environment variables
    nodes: List     = field(default_factory = list)   # nodes
    units: Dict     = field(default_factory = dict)   # custom units
    sources: Dict   = field(default_factory = dict)   # custom reference sources
    functions: Dict = field(default_factory = dict)  # custom native functions

    # Reference on the current node
    autoref: str = None

    # Documentation mode
    docs: bool = False  

    # Hierarchy list
    parent_indents: List[int] = field(default_factory = list)  # indent level
    parent_names: List[str]   = field(default_factory = list)    # list of parent names

    # Case list
    case_counts: List[int]  = field(default_factory = list)     # case number in a current condition
    case_names: List[str]   = field(default_factory = list)    # condition case name
    
    def __post_init__(self):
        self.parent_indents.append(-1)
        self.case_counts.append(0)
        self.case_names.append('')
    
    def copy(self):
        """ Copy a new object from self
        """
        return copy.copy(self)
        
    def pop_node(self):
        """ Pop first node out of the node list
        """
        return self.nodes.pop(0)

    def prepend_nodes(self, nodes):
        """ Prepend new nodes to the node list
        """
        self.nodes = nodes + self.nodes
    
    def update_hierarchy(self, node, excluded):
        """ Register new node to the hierarchy

        :param node: Node that should be added
        :param list excluded: List of node keywords that should be excluded from the hierarchy
        """
        if node.name is not None and node.keyword not in excluded:
            while node.indent<=self.parent_indents[-1]:
                self.parent_indents.pop()
                self.parent_names.pop()
            self.parent_names.append(node.name)
            self.parent_indents.append(node.indent)
            node.name = Sign.SEPARATOR.join(self.parent_names)

    def is_case(self):
        """ Checks if outside or inside of a valid case clause
        """
        return not self.case_names[-1] or self.case_counts[-1]<2

    def _end_case(self):
        self.case_names.pop()
        self.case_counts.pop()
    
    def solve_case(self, node):
        """ Manage condition nodes

        :param node: Condition node
        """
        casename = self.case_names[-1]
        if node.name.endswith(Sign.CONDITION + Keyword.CASE):
            if casename+Keyword.CASE!=node.name:   # register new case
                self.case_names.append(node.name[:-4])
                self.case_counts.append(0)
            if node.value or self.case_counts[-1]==1:
                self.case_counts[-1] += 1
        elif node.name==casename + Keyword.ELSE:
            self.case_counts[-1] += 1
        elif node.name==casename + Keyword.END:    # end case using a keyword
            self._end_case()
        else:
            raise Exception(f"Invalid condition:", node.name)

    def prepare_node(self, node):
        """ Manage parameter nodes in a condition

        :param node: Parameter node
        """
        if not self.case_names[-1]: # outside of any condition
            return        
        if not node.name.startswith(self.case_names[-1]): # ending case at lower indent
            self._end_case()
        node.name = node.name.replace(
            Sign.CONDITION + Keyword.CASE + Sign.SEPARATOR,''
        )
        node.name = node.name.replace(
            Sign.CONDITION + Keyword.ELSE + Sign.SEPARATOR,''
        )

    
    def add_unit(self, name:str, unit:Quantity):
        """ Add a new source

        :param str name: Name of a new unit
        :param Quantity unit: Quantity object
        :param float num: Numerical value
        :param str expr: Unit expression
        """
        name = f"[{name}]"
        if name in self.units:
            raise Exception("Reference unit alread exists:", name)
        self.units[name] = {'magnitude':unit.magnitude.value, 'dimensions':unit.baseunits.dimensions.value()} #,'prefixes':['k','M','G']}
        
    def add_source(self, name:str, source:str, path:str):
        """ Add a new source

        :param str name: Name of a new source
        :param source: Source can be either a DIP object or a text
        :param str path: Source path
        """
        if name in self.sources:
            raise Exception("Reference source alread exists:", name)
        self.sources[name] = EnvSource(source=source, path=str(path), name=name)
        
    def query(self, query:str, namespace:int=Namespace.NODES):
        """ Select local nodes according to a query

        :param str query: Node selection query
        :param str namespace: Query namespace (nodes, sources, or units)
        """
        if namespace==Namespace.NODES:
            nodes = []
            if query==Sign.WILDCARD:
                return [node.copy() for node in self.nodes]
            elif query[-2:]==Sign.SEPARATOR + Sign.WILDCARD:
                for node in self.nodes:
                    if node.name.startswith(query[:-1]):
                        node = node.copy()
                        node.name = node.name[len(query[:-1]):]
                        nodes.append(node.copy())
            else:
                for node in self.nodes:
                    if node.name==query:
                        node = node.copy()
                        node.name = node.name.split(Sign.SEPARATOR)[-1]
                        nodes.append(node.copy())
            return nodes
        elif namespace==Namespace.SOURCES:
            if query==Sign.WILDCARD:   # return all sources
                return self.sources
            else:                     # return particular source
                if query not in self.sources:
                    raise Exception("Requested source does not exists:", query)
                return {query: self.sources[query]}
        elif namespace==Namespace.UNITS:
            if query==Sign.WILDCARD:   # return all units
                return self.units
            else:                     # return particular unit
                if query not in self.units:
                    raise Exception("Requested unit does not exists:", query)
                return {query: self.units[query]}
        else:
            raise Exception("Invalid query namespace selected:", namespace)
        
    def request(self, path:str, count:int=None, namespace:str=Namespace.NODES):
        """ Request nodes from a path

        :param str path: Request path
        :param int count: Number of nodes that should be selected
        :param str namespace: Query namespace (nodes, sources, or units)
        """
        if self.autoref and path == Sign.QUERY: # reference type {?}
            filename,query = '', self.autoref
        elif Sign.QUERY in path:                    # reference type {source?query}
            filename,query = path.split(Sign.QUERY)
        else:                                      # reference type {source}
            filename,query = path,Sign.WILDCARD
        if filename:  # use external source to parse the values
            source = self.sources[filename].source
            if isinstance(source, str):
                return source
            else:
                nodes = source.env.query(query, namespace)
        else:         # use values parsed in the current file
            if not self.nodes:
                raise Exception(f"Local nodes are not available for DIP import:", path)
            nodes = self.query(query, namespace)
        if count:
            if isinstance(count, list) and len(nodes) not in count:
                raise Exception(f"Path returned invalid number of nodes:", path, count, len(nodes))
            elif np.isscalar(count) and len(nodes)!=count:
                raise Exception(f"Path returned invalid number of nodes:", path, count, len(nodes))
        return nodes

    def data(self, format=Format.VALUE, verbose=False):
        """ Return parsed values as a dictionary

        :param bool verbose: Display node values
        :param str format: Return data as values only, DIP datatypes, or tuples
        """
        data = {}
        for node in self.nodes:
            if format==Format.VALUE:
                data[node.name] = node.value.value
            elif format==Format.TYPE:
                data[node.name] = node.value
            elif format==Format.TUPLE:
                if isinstance(node.value, NumberType) and node.value.unit is not None:
                    data[node.name] = (node.value.value, node.value.unit)
                else:
                    data[node.name] = node.value.value
            elif format==Format.QUANTITY:
                if isinstance(node.value, NumberType):
                    data[node.name] = Quantity(node.value.value, node.value.unit)
                else:
                    data[node.name] = node.value.value
            elif format==Format.NODE:
                data[node.name] = node
            else:
                raise Exception("Data format not recognized:", format)
            if verbose:
                 print(node.name,'|',node.indent,'|',node.keyword,'|',str(node.value),
                       '|',repr(node.units_raw), end='')
                 if hasattr(node,'options'):
                     if node.options:
                         print(' |',node.options, end='') 
                 print()
        return data



class Numeric:
    PRECISION   = 1e-6

class Keyword:
    TRUE        = 'true'
    FALSE       = 'false'
    CASE        = 'case'
    ELSE        = 'else'
    END         = 'end'
    UNIT        = 'unit'
    SOURCE      = 'source'
    OPTIONS     = 'options'
    CONSTANT    = 'constant'
    FORMAT      = 'format'
    CONDITION   = 'condition'

class Sign:
    QUERY       = '?'
    WILDCARD    = '*'
    NEGATE      = '~'
    DEFINED     = '!'
    SEPARATOR   = '.'
    CONDITION   = '@'
    VARIABLE    = '$'
    VALIDATION  = '!'

class Namespace:
    NODES       = 1
    UNITS       = 2
    SOURCES     = 3
    
class Format:
    VALUE       = 1
    TUPLE       = 2
    TYPE        = 3
    NODE        = 4
    QUANTITY    = 5



import numpy as np
import pytest
from math import isclose
import os
import sys
import time
sys.path.insert(0, 'src')

from scinumtools.units import *

def test_scalars():
    q = Quantity(23, 'erg')
    
    tstart = time.time()
    for i in range(100):
        q.to('keV')
    ttotal = time.time() - tstart
    assert ttotal < 3.e-1
    
def test_arrays():
    q = Quantity(np.linspace(1,100,40), 'cm')
    
    tstart = time.time()
    for i in range(100):
        r = q.to('au')
    ttotal = time.time() - tstart
    assert ttotal < 3.e-1
    
    tstart = time.time()
    for i in q:
        r = i
    ttotal = time.time() - tstart
    assert ttotal < 3.e-1
    


import numpy as np
import pandas as pd
from textwrap import dedent
import sys
sys.path.insert(0, 'src')

import scinumtools as snt

def test_progressbar():

    nsteps = 200
    pb = snt.ProgressBar(nsteps)
    for i in range(nsteps):
        pb.step()
    pb.close()

    with snt.ProgressBar(nsteps) as pb:
        pb.step()
        


import numpy as np
import pytest
from math import isclose
import os
import sys
sys.path.insert(0, 'src')

from scinumtools.units import *
from scinumtools.units.settings import SI
from scinumtools.units.unit_environment import *
    
def test_unit_list():
    
    assert str(Unit())
    assert str(Constant())

    #print(Constant())
    #print(Unit())
    #exit(1)
    
def test_unique_unique():
    
    assert check_unique_symbols()

def test_custom_units():
    
    # register custom unit
    units = {
        'x': {'magnitude':3, 'dimensions':[3,2,-1,0,0,1,0,0]},   # defined as a dictionary
        'y': Quantity(2, 'cm/g2')                                # defined as a quantity
    }
    with UnitEnvironment(units):
    
        q = Quantity(1, 'x')
        assert str(q) == "Quantity(1.000e+00 x)"
        assert q.baseunits.magnitude          == 3.0
        assert q.baseunits.dimensions.value() == [3,2,-1,0,0,1,0,0]
        
        q = Quantity(1, 'y')
        assert str(q) == "Quantity(1.000e+00 y)"
        assert q.baseunits.magnitude          == 0.02
        assert q.baseunits.dimensions.value() == [1,-2,0,0,0,0,0,0]

        with pytest.raises(Exception) as excinfo:
            Quantity(1, 'kx')
        assert excinfo.value.args[0]=="Unit cannot have any prefixes:"
        assert excinfo.value.args[1]=="x"

        # try to register new unit with already existing symbol
        with pytest.raises(Exception) as excinfo:
            with UnitEnvironment(units):
                pass
        assert excinfo.value.args[0]=="Unit with this symbol already exists:"
        assert excinfo.value.args[1]=="x"
    
    # make sure that units does not exist outside of the environment
    assert 'x' not in UNIT_STANDARD
    
    # register new unit with a custom converter 
    class CustomUnitType(UnitType):
        def _istype(self):
            return False
    units = {'x': {'magnitude':3, 'dimensions':[3,2,-1,0,0,1,0,0], 'definition':CustomUnitType}}
    env = UnitEnvironment(units)
    assert 'x' in UNIT_STANDARD
    assert CustomUnitType in UNIT_TYPES
    env.close()
    
    # make sure that units and types does not exist aftr environment is closed
    assert 'x' not in UNIT_STANDARD
    assert CustomUnitType not in UNIT_TYPES
    

def test_prefixes():
    
    # unit can have any prefix
    assert str(Quantity(1, 'km'))    == "Quantity(1.000e+00 km)"
    
    # unit cannot have prefixes
    with pytest.raises(Exception) as excinfo:
        Quantity(1, 'kCel')
    assert excinfo.value.args[0]=="Unit cannot have any prefixes:"
    assert excinfo.value.args[1]=="Cel"
    
    # unit can have only selected prefixes
    with pytest.raises(Exception) as excinfo:
        Quantity(1, 'mpc')
    assert excinfo.value.args[0]=="Unit can have only following prefixes:"
    assert excinfo.value.args[1]==['k', 'M', 'G', 'T']
    

def test_definitions():
    
    for symbol, unit in UNIT_STANDARD.items():
        if not isinstance(unit.definition,str):
            continue
        q = Quantity(1, unit.definition)
        base = q.baseunits
        magnitude = q.magnitude.value*base.magnitude
        if base.dimensions.value(dtype=list) != unit.dimensions or not isclose(magnitude,  unit.magnitude, rel_tol=MAGNITUDE_PRECISION):
            print(q, symbol, unit.definition, unit.magnitude)
        assert isclose(magnitude,  unit.magnitude, rel_tol=MAGNITUDE_PRECISION)
        assert base.dimensions.value(dtype=list) == unit.dimensions
        
    for symbol, unit in UNIT_PREFIXES.items():
        if not isinstance(unit.definition,str):
            continue
        q = Quantity(1, unit.definition)
        base = q.baseunits
        assert isclose(q.magnitude.value,  unit.magnitude, rel_tol=MAGNITUDE_PRECISION)
        assert base.dimensions.value(dtype=list) == unit.dimensions
        
def test_quantities():
    
    assert SI.Pressure.value       == '#SPRE'
    assert CGS.Energy.value        == '#CENE'
    assert AU.ElectricCharge.value == '#AECH'
    assert type(SI.Pressure)       == SI

    text = [
        "#############################################",
        "# Do not modify this file!                  #",
        "# It is generated automatically in:         #",
        "# tests/units/test_list.py::test_quantities #",
        "#############################################",
        ""
        "QUANTITY_UNITS = {",
    ]
    symbols = []
    for q in range(len(QUANTITY_LIST)):
        name = QUANTITY_LIST.name[q]
        text.append(f"  # {name}")
        for system in ['SI','CGS','AU']:
            definition = getattr(QUANTITY_LIST,system)[q]
            if definition is None:
                continue
            else:
                symbol = QUANTITY_LIST.symbol[q]
                unitid = f"#{system[0]}{symbol}"
                atom = UnitSolver(definition)
                base = BaseUnits(atom.baseunits)
                c1 = f"{atom.magnitude*base.magnitude},"
                c2 = f"{base.dimensions.value()}"
                symbols.append(unitid)
                text.append(f"  '{unitid}': ({c1:25s}{c2:27s}), # {definition}") 
    text.append("}")
    text = "\n".join(text)
    
    # test if new symbols are unique
    assert len(np.unique(symbols)) == len(symbols)
    
    # test if we produced a valid Python code
    exec(text)
    assert QUANTITY_UNITS

    # test if quantity lists work
    assert str(Quantity(23, SI.Pressure))          == 'Quantity(2.300e+01 #SPRE)'
    assert str(Quantity(23, 'Pa').to(SI.Pressure)) == 'Quantity(2.300e+01 #SPRE)'
    assert str(Quantity(23, SI.Pressure).to('Pa')) == 'Quantity(2.300e+01 Pa)'
    assert str(Quantity(1, '[a_0]').to(AU.Length)) == 'Quantity(1.000e+00 #ALEN)'
    assert str(Unit(AU.Length))                    == "Quantity(1.000e+00 #ALEN)"
    assert str(Unit('[a_0]')/Unit(AU.Length))      == 'Quantity(1.000e+00)'
    assert str(Unit(AU.Length)*Unit('s'))          == 'Quantity(1.000e+00 #ALEN*s)'
    assert str(Quantity(1, '#ALEN/s').to('m/s'))   == 'Quantity(5.292e-11 m*s-1)'
    
    # save the new version of the code
    path_units = "src/scinumtools/units"
    assert os.path.isdir(path_units)
    with open(f'{path_units}/unit_list.py','w') as f:
        f.write(text)



from .node_base import BaseNode
from .node_integer import IntegerNode
from .node_float import FloatNode
from .node_string import StringNode

class OptionNode(BaseNode):
    keyword: str = 'option'

    @staticmethod
    def is_node(parser):
        parser.part_equal()    
        if parser.is_parsed('part_equal'):
            parser.part_value()
            parser.part_units()
            parser.part_comment()
            return OptionNode(parser)
        parser.kwd_options()
        if parser.is_parsed('kwd_options'):
            parser.part_value()
            parser.part_units()
            parser.part_comment()
            return OptionNode(parser)
        
    def parse(self, env):
        node = env.nodes[-1]
        if not isinstance(node,(IntegerNode,FloatNode,StringNode)):
            raise Exception(f"Node '{env.nodes[-1].keyword}' does not support options")
        if self.dimension:
            for value in self.cast_value():
                self.value_raw = value
                env.nodes[-1].set_option(self.copy(), env)
        else:
            env.nodes[-1].set_option(self, env)
        return None



from typing import List
import numpy as np
from decimal import Decimal
import re
from typing import Union

from .settings import *
from .unit_types import *
from .magnitude import Magnitude
from .dimensions import Dimensions
from .base_units import BaseUnits, get_unit_base
from .fraction import Fraction
from .unit_solver import UnitSolver

HANDLED_FUNCTIONS = {}

class Quantity:
    
    magnitude: Magnitude      # magnitude
    baseunits: BaseUnits      # base units

    def __init__(
        self, 
        magnitude: Union[int,float,Decimal,list,np.ndarray,Magnitude],
        baseunits: Union[str,list,np.ndarray,Dimensions,dict,BaseUnits] = None,
        abse: Union[int,float] = None,
        rele: Union[int,float] = None
    ):
        # Set magnitude
        if isinstance(magnitude, (int,float,Decimal,list,np.ndarray)):
            self.magnitude = Magnitude(magnitude, abse=abse, rele=rele)
        elif isinstance(magnitude, Magnitude):
            if abse is not None:
                magnitude.abse(abse)
            elif rele is not None:
                magnitude.rele(rele)
            self.magnitude = magnitude
        else:
            raise Exception("Magnitude can be either a number or an list/array of numbers", magnitude)
        # Set base units
        if baseunits is None or isinstance(baseunits, (dict,Dimensions,list,np.ndarray)):
            self.baseunits = BaseUnits(baseunits)
        elif isinstance(baseunits, BaseUnits):
            self.baseunits = baseunits
        elif isinstance(baseunits, (SI,CGS,AU)):
            self.baseunits = BaseUnits(baseunits)
        elif isinstance(baseunits, str):
            atom = UnitSolver(baseunits)
            self.magnitude *= atom.magnitude
            self.baseunits = BaseUnits(atom.baseunits)
        elif isinstance(baseunits, Quantity):
            self.magnitude *= baseunits.magnitude
            self.baseunits = baseunits.baseunits
        else:
            raise Exception("Insufficient quantity definition", magnitude, baseunits)
        # rebase if dimensions are zero
        if self.baseunits.dimensions.nodim:
            baseunits = {}
            for unitid, exp in self.baseunits.baseunits.items():
                base = get_unit_base(unitid, exp)
                if base.dimensions.nodim:
                    baseunits[unitid] = exp
                    continue
                else:
                    self.magnitude *= base.magnitude
            self.baseunits = BaseUnits(baseunits)

    def _add(self, left, right):
        for utype in UNIT_TYPES:
            if c := utype(left.baseunits, right.baseunits):
                magnitude = c.add(left, right)
                baseunits = left.baseunits
                return Quantity(magnitude, baseunits)
        else:
            raise Exception("Unsupported addition between units:", left, right)

    def __add__(self, other):
        if not isinstance(other, Quantity):
            other = Quantity(other)
        return self._add(self, other)
    
    def __radd__(self, other):
        if not isinstance(other, Quantity):
            other = Quantity(other)
        return self._add(other, self)

    def _sub(self, left, right):
        for utype in UNIT_TYPES: 
            if c := utype(left.baseunits, right.baseunits):
                magnitude = c.sub(left, right)
                baseunits = left.baseunits
                return Quantity(magnitude, baseunits)
        else:
            raise Exception("Unsupported subtraction between units:", left, right)

    def __sub__(self, other):
        if not isinstance(other, Quantity):
            other = Quantity(other)
        return self._sub(self, other)

    def __rsub__(self, other):
        if not isinstance(other, Quantity):
            other = Quantity(other)
        return self._sub(other, self)
    
    def _mul(self, left, right):
        magnitude = left.magnitude * right.magnitude
        baseunits = left.baseunits + right.baseunits
        return Quantity(magnitude, baseunits)

    def __mul__(self, other):
        if not isinstance(other, Quantity):
            other = Quantity(other)
        return self._mul(self, other)

    def __rmul__(self, other):
        if not isinstance(other, Quantity):
            other = Quantity(other)
        return self._mul(other, self)
    
    def _truediv(self, left, right):
        magnitude = left.magnitude / right.magnitude
        baseunits = left.baseunits - right.baseunits
        return Quantity(magnitude, baseunits)

    def __truediv__(self, other):
        if not isinstance(other, Quantity):
            other = Quantity(other)
        return self._truediv(self, other)
    
    def __rtruediv__(self, other):
        if not isinstance(other, Quantity):
            other = Quantity(other)
        return self._truediv(other, self)
    
    def __pow__(self, power: Union[float,int,tuple,Fraction]):
        if isinstance(power, tuple):
            exp = power[0]/power[1]
        elif isinstance(power, Fraction):
            exp = power.value(dtype=float)
        else:
            exp = power
        magnitude = self.magnitude**exp
        baseunits = self.baseunits*power
        return Quantity(magnitude, baseunits)

    def __neg__(self):
        return Quantity(-self.magnitude, self.baseunits)
    
    def __eq__(self, other):
        if isinstance(other, (int, float)):
            other = Quantity(other)
        if not np.allclose(self.magnitude.value, other.magnitude.value, rtol=MAGNITUDE_PRECISION):
            return False
        if not self.baseunits==other.baseunits:
            return False
        return True
    
    def __str__(self):
        magnitude = str(self.magnitude)
        baseunits = self.baseunits.expression
        if baseunits:
            return f"Quantity({magnitude:s} {baseunits})"
        else:
            return f"Quantity({magnitude:s})"
            
    def __repr__(self):
        magnitude = str(self.magnitude)
        baseunits = self.baseunits.expression
        if baseunits:
            return f"Quantity({magnitude:s} {baseunits})"
        else:
            return f"Quantity({magnitude:s})"

    def __getitem__(self, key):
        return Quantity(self.magnitude.value[key], self.baseunits)
        
    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        if ufunc==np.sqrt:
            return Quantity(ufunc(inputs[0].magnitude.value), inputs[0].baseunits/2)
        elif ufunc==np.cbrt:
            return Quantity(ufunc(inputs[0].magnitude.value), inputs[0].baseunits/3)
        elif ufunc==np.power:
            return Quantity(ufunc(inputs[0].magnitude.value,inputs[1]), inputs[0].baseunits*inputs[1])
        elif ufunc in [np.sin, np.cos, np.tan]:
            return Quantity(ufunc(inputs[0].to('rad').magnitude.value))
        elif ufunc in [np.arcsin, np.arccos, np.arctan]:
            return Quantity(ufunc(inputs[0].to(None).magnitude.value),'rad')
        elif ufunc in [np.isnan, np.isnat]:
            return ufunc(inputs[0].magnitude.value)
        else:
            return Quantity(ufunc(inputs[0].magnitude.value), inputs[0].baseunits)
    
    def __array_function__(self, func, types, args, kwargs):
        if func not in HANDLED_FUNCTIONS:
            raise NotImplementedError()
        return HANDLED_FUNCTIONS[func](*args, **kwargs)
    
    def _convert(self, magnitude1, baseunits1, baseunits2):
        for utype in UNIT_TYPES:
            if c := utype(baseunits1, baseunits2):
                return c.convert(magnitude1)
        else:
            raise Exception("Unsupported conversion between units:", baseunits1.expression, baseunits2.expression)

    def value(self, expression=None, dtype=None):
        if expression:
            value = self._convert(self.magnitude, self.baseunits, BaseUnits(expression)).value
        else:
            value = self.magnitude.value
        if dtype:
            return value.astype(dtype) if isinstance(value, np.ndarray) else dtype(value)
        else:
            return value

    def units(self):
        return self.baseunits.expression

    def to(self, units: Union[str,list,np.ndarray,Dimensions,dict,BaseUnits]):
        if isinstance(units, Quantity):
            baseunits = units.baseunits
            self.magnitude = self._convert(self.magnitude, self.baseunits, baseunits) / units.magnitude
        else:
            baseunits = BaseUnits(units)
            self.magnitude = self._convert(self.magnitude, self.baseunits, baseunits)
        self.baseunits = baseunits
        return self
        
    def abse(self, error: Union[int,float] = None):
        if error is None:
            return self.magnitude.abse()
        else:
            self.magnitude.abse(error)
            return self
        
    def rele(self, error: Union[int,float] = None):
        if error is None:
            return self.magnitude.rele()
        else:
            self.magnitude.rele(error)
            return self
    
    def rebase(self):
        factor = 1
        baseunits = {}
        for unitid1,exp1 in self.baseunits.baseunits.items():
            # find base units
            base1 = get_unit_base(unitid1)
            dim1 = str(base1.dimensions.value(dtype=tuple))
            if dim1 in baseunits:
                # exists: convert units
                base0 = get_unit_base(baseunits[dim1][0])
                factor *= (base1.magnitude/base0.magnitude)**exp1.value(dtype=float)
                baseunits[dim1][1] += exp1
            else:
                # does not exist: register new
                baseunits[dim1] = [unitid1,exp1] 
        # construct new base units
        self.magnitude *= factor
        self.baseunits = BaseUnits({unitid:exp for unitid,exp in baseunits.values()})
        return self

def implements(np_function):
    def decorator(func):
        HANDLED_FUNCTIONS[np_function] = func
        return func
    return decorator

@implements(np.linspace)
def linspace(a, b, c, **kwargs):
    if isinstance(a,Quantity):
        b = b.to(a.baseunits) if isinstance(b,Quantity) else Quantity(b, a.baseunits)
    else:
        a = a.to(b.baseunits) if isinstance(a,Quantity) else Quantity(a, b.baseunits)
    return Quantity(np.linspace(a.magnitude.value, b.magnitude.value, c, **kwargs), a.baseunits)

@implements(np.logspace)
def logspace(a, b, c, **kwargs):
    if isinstance(a,Quantity):
        b = b.to(a.baseunits) if isinstance(b,Quantity) else Quantity(b, a.baseunits)
    else:
        a = a.to(b.baseunits) if isinstance(a,Quantity) else Quantity(a, b.baseunits)
    return Quantity(np.logspace(a.magnitude.value, b.magnitude.value, c, **kwargs), a.baseunits)

@implements(np.absolute)
def absolute(a, **kwargs):
    return Quantity(np.absolute(a.magnitude.value), a.baseunits)

@implements(np.abs)
def abs(a, **kwargs):
    return Quantity(np.abs(a.magnitude.value), a.baseunits)

@implements(np.round)
def round(a, **kwargs):
    return Quantity(np.round(a.magnitude.value), a.baseunits)

@implements(np.floor)
def floor(a, **kwargs):
    return Quantity(np.floor(a.magnitude.value), a.baseunits)

@implements(np.ceil)
def ceil(a, **kwargs):
    return Quantity(np.ceil(a.magnitude.value), a.baseunits)
    
@implements(np.iscomplexobj)
def iscomplexobj(a, **kwargs):
    return False



import numpy as np
from dataclasses import dataclass, field
from decimal import Decimal
from typing import Union

class Magnitude:
    value: Union[int,float,Decimal,np.ndarray]
    error: Union[int,float,Decimal,np.ndarray] = None   # absolute error 

    def __init__(self, value: float, abse: float = None, rele: float = None):
        # set value
        if isinstance(value, (float,int)):
            self.value = float(value)
        elif isinstance(value, Decimal):
            self.value = value
        elif isinstance(value, list):
            self.value = np.array(value, dtype=float)
        elif isinstance(value, np.ndarray):
            self.value = value.astype(float)
        else:
            raise Exception("Magnitude value can be either a number or an list/array of numbers", value)
        # set error
        if abse is not None and rele is not None:
            raise Exception("Magnitude cannot have both absolute and relative errors!", abse, rele)
        elif abse is not None and rele is None:
            self.error = abse
        elif abse is None and rele is not None:
            self.error = self._rel_to_abs(rele)
        # set array of errors if value is an array
        if isinstance(self.value, np.ndarray) and self.error is not None:
            self.error = np.full_like(self.value, self.error)
            
    def _rel_to_abs(self, rele):
        return self.value*rele/100
        
    def _abs_to_rel(self, abse=None):
        if abse is None:
            abse = self.error
        return 100*abse/self.value
        
    @staticmethod
    def parse_string(value, error):
        def formatter(val, err):
            exps = np.floor(np.log10([np.abs(val),err]))
            val = val*10**-exps[0]
            ndec = int(np.abs(exps[0]-exps[1])+1)
            vformat = f".0{ndec}f"
            err = int(np.round(np.round(err*10**(1-exps[1]),decimals=1))) # double round because of values like 0.4999999
            exponent = int(np.abs(np.round(exps[0])))
            sign = "+" if exps[0]>=0  else "-"
            return f"{val:{vformat}}({err:2d})e{sign}{exponent:-02d}"
        if isinstance(value, np.ndarray):
            out = np.empty_like(value, dtype=object)
            for index, x in np.ndenumerate(out):
                out[index] = formatter(value[index],error[index])
            return np.array2string(out, formatter={'all': lambda x: str(x)})
        else:
            return formatter(value,error)
    
    def _str(self):    
        if self.error is None:
            if isinstance(self.value, (list,np.ndarray)):
                with np.printoptions(precision=3, suppress=False, threshold=5):
                    return str(self.value)
            else:
                return f"{self.value:.03e}"
        else:          
            return Magnitude.parse_string(self.value, self.error)
    
    def __str__(self):
        return self._str()
        
    def __repr__(self):
        return self._str()

    def _add(self, left, right):
        if isinstance(left.value, Decimal) or isinstance(right.value, Decimal):
            value = Decimal(left.value) + Decimal(right.value)
        else:
            value = left.value + right.value
        if left.error is None and right.error is None:
            error = None
        elif left.error is None and right.error is not None:
            error = right.error
        elif left.error is not None and right.error is None:
            error = left.error
        else:
            error = left.error + right.error
        return Magnitude(value, error)
        
    def __add__(self, other):
        if not isinstance(other, Magnitude):
            other = Magnitude(other)
        return self._add(self, other)
        
    def __radd__(self, other):
        if not isinstance(other, Magnitude):
            other = Magnitude(other)
        return self._add(other, self)
        
    def _sub(self, left, right):
        if isinstance(left.value, Decimal) or isinstance(right.value, Decimal):
            value = Decimal(left.value) - Decimal(right.value)
        else:
            value = left.value - right.value
        if left.error is None and right.error is None:
            error = None
        elif left.error is None and right.error is not None:
            error = right.error
        elif left.error is not None and right.error is None:
            error = left.error
        else:
            error = left.error + right.error
        return Magnitude(value, error)
        
    def __sub__(self, other):
        if not isinstance(other, Magnitude):
            other = Magnitude(other)
        return self._sub(self, other)
        
    def __rsub__(self, other):
        if not isinstance(other, Magnitude):
            other = Magnitude(other)
        return self._sub(other, self)
        
    def _mul(self, left, right):
        if isinstance(left.value, Decimal) or isinstance(right.value, Decimal):
            value = Decimal(left.value) * Decimal(right.value)
        else:
            value = left.value * right.value
        if left.error is None and right.error is None:
            error = None
        elif left.error is None and right.error is not None:
            error = right.error * left.value
        elif left.error is not None and right.error is None:
            error = left.error * right.value
        else:
            maxerror = np.abs((left.value+left.error)*(right.value+right.error) - value)
            minerror = np.abs((left.value-left.error)*(right.value-right.error) - value)
            error = np.max([maxerror,minerror])
        return Magnitude(value, error)
        
    def __mul__(self, other):
        if not isinstance(other, Magnitude):
            other = Magnitude(other)
        return self._mul(self, other)
        
    def __rmul__(self, other):
        if not isinstance(other, Magnitude):
            other = Magnitude(other)
        return self._mul(other, self)
        
    def _truediv(self, left, right):
        if isinstance(left.value, Decimal) or isinstance(right.value, Decimal):
            value = Decimal(left.value) / Decimal(right.value)
        else:
            value = left.value / right.value
        if left.error is None and right.error is None:
            error = None
        elif left.error is None and right.error is not None:
            maxerror = np.abs(left.value / (right.value+right.error) - value)
            minerror = np.abs(left.value / (right.value-right.error) - value)
            error = np.max([maxerror,minerror])
        elif left.error is not None and right.error is None:
            error = left.error / right.value
        else:
            maxerror = np.abs((left.value+left.error)/(right.value-right.error) - value)
            minerror = np.abs((left.value-left.error)/(right.value+right.error) - value)
            error = np.max([maxerror,minerror])
        return Magnitude(value, error)
        
    def __truediv__(self, other):
        if not isinstance(other, Magnitude):
            other = Magnitude(other)
        return self._truediv(self, other)
        
    def __rtruediv__(self, other):
        if not isinstance(other, Magnitude):
            other = Magnitude(other)
        return self._truediv(other, self)
        
    def __pow__(self, power: Union[float,int]):
        value = self.value**power
        if self.error is not None:
            error = self._rel_to_abs(self._abs_to_rel()*power)
        else:
            error = None
        return Magnitude(value, error)
        
    def __neg__(self):
        return Magnitude(-self.value, self.error)
        
    def abse(self, abse=None):
        if abse is None:
            return self.error
        else:
            self.error = abse
            return self
        
    def rele(self, rele=None):
        if rele is None:
            return self._abs_to_rel()
        else:
            self.error = self._rel_to_abs(rele)
            return self
            
            


import numpy as np
from dataclasses import dataclass, field, fields
from typing import Union

from .settings import *
from .fraction import Fraction
from .dimensions import Dimensions
from .unit_solver import UnitSolver

@dataclass
class Base:
    
    magnitude: float
    dimensions: Dimensions
    units: str
    expression: str

def get_unit_base(unitid: str, exp: Fraction = None):
    if exp is None:
        exp = Fraction(1)
    if unitid.startswith(SYMBOL_SYSTEM_UNIT):
        prefix, base = '', unitid
        qu = QUANTITY_UNITS[unitid]
        magnitude  = qu[0] ** exp.value(dtype=float)
        dimensions = Dimensions(*qu[1])*exp
    elif SYMBOL_UNITID in unitid:
        prefix, base = unitid.split(SYMBOL_UNITID)
        magnitude  = (UNIT_PREFIXES[prefix].magnitude*UNIT_STANDARD[base].magnitude) ** exp.value(dtype=float)
        dimensions = Dimensions(*UNIT_STANDARD[base].dimensions)*exp
    else:
        prefix, base = '', unitid
        magnitude  = UNIT_STANDARD[base].magnitude ** exp.value(dtype=float)
        dimensions = Dimensions(*UNIT_STANDARD[base].dimensions)*exp
    if exp.num==1 and exp.den==1:
        expression = f"{prefix}{base}"
    else:
        expression = f"{prefix}{base}{exp}"
    return Base(magnitude, dimensions, base, expression)

class BaseUnits:

    baseunits: dict
    magnitude: float
    dimensions: Dimensions
    units: list
    expression: Union[str,list]
    nodim: bool
    nobase: bool

    def __init__(self, baseunits: Union[str,list,dict,Dimensions]=None):
        if baseunits is None:
            self.baseunits = {}
        elif isinstance(baseunits, dict):
            self.baseunits = baseunits
        elif isinstance(baseunits, Dimensions):
            self.baseunits = baseunits.value(dtype=dict)
        elif isinstance(baseunits, (list, np.ndarray)):
            self.baseunits = Dimensions(*baseunits).value(dtype=dict)
        elif isinstance(baseunits, BaseUnits):
            self.baseunits = baseunits.baseunits
        elif isinstance(baseunits, (SI, CGS, AU)):
            self.baseunits = {baseunits.value: 1}
        elif isinstance(baseunits, str):
            self.baseunits = UnitSolver(baseunits).baseunits
        else:
            raise Exception("Cannot initialize BaseUnits with given argument:", baseunits)
        # calculate total base
        self.magnitude = 1
        self.dimensions = Dimensions()
        self.units = []
        self.expression = []
        self.nobase = True
        for unitid in list(self.baseunits.keys()):
            if not isinstance(self.baseunits[unitid], Fraction):
                self.baseunits[unitid] = Fraction(self.baseunits[unitid])
            if self.baseunits[unitid].num==0:
                del self.baseunits[unitid]
                continue
            self.nobase = False
            ubase = get_unit_base(unitid, self.baseunits[unitid])
            self.magnitude *= ubase.magnitude
            self.dimensions += ubase.dimensions
            self.units.append(ubase.units)
            self.expression.append(ubase.expression)
        self.expression = SYMBOL_MULTIPLY.join(self.expression) if self.expression else None
        self.nodim = self.dimensions.nodim

    def __str__(self):
        baseunits = []
        for unit,exp in self.baseunits.items():
            if exp.num not in [0, -0]:
                unit = unit.replace(SYMBOL_UNITID,"")
                baseunits.append(f"{unit}={str(exp)}")
        baseunits = " ".join(baseunits)
        return f"BaseUnits({baseunits})"

    def __repr__(self):
        baseunits = []
        for unit,exp in self.baseunits.items():
            if exp.num not in [0, -0]:
                unit = unit.replace(SYMBOL_UNITID,"")
                baseunits.append(f"{unit}={str(exp)}")
        baseunits = " ".join(baseunits)
        return f"BaseUnits({baseunits})"

    def __add__(self, other):
        baseunits = dict(self.baseunits)
        for unit,exp in other.baseunits.items():
            baseunits[unit] = baseunits[unit]+exp if unit in baseunits else exp
        return BaseUnits(baseunits)
    
    def __sub__(self, other):
        baseunits = dict(self.baseunits)
        for unit,exp in other.baseunits.items():
            baseunits[unit] = baseunits[unit]-exp if unit in baseunits else -exp
        return BaseUnits(baseunits)

    def __mul__(self, other):
        baseunits = dict(self.baseunits)
        for unit,exp in self.baseunits.items():
            baseunits[unit] *= other
        return BaseUnits(baseunits)

    def __truediv__(self, div):
        baseunits = dict(self.baseunits)
        for unit,exp in self.baseunits.items():
            baseunits[unit] /= div
        return BaseUnits(baseunits)
    
    def __eq__(self, other):
        if len(self.baseunits) != len(other.baseunits):
            return False
        for unit,exp in self.baseunits.items():
            if unit not in other.baseunits:
                return False
            elif not other.baseunits[unit] == exp:
                return False
        return True
    
    def value(self):
        """ Return base units as a plain dictionary
        """
        baseunits = {}
        for unitid,exp in self.baseunits.items():
            baseunits[unitid] = exp.value()
        return baseunits



import numpy as np
import pytest
from math import isclose
import os
import sys
sys.path.insert(0, 'src')

from scinumtools.units import *
from scinumtools.units.unit_solver import AtomParser, Atom

def test_atom():
    
    a = Atom(3e2, {'k:g':Fraction(2),'m':Fraction(3,2)})
    assert str(a)    == "Atom(3.000e+02 kg=2 m=3:2)"
    assert str(a*a)  == "Atom(9.000e+04 kg=4 m=3)"
    assert str(a/a)  == "Atom(1.000e+00)"

def test_atom_parser():
    
    assert str(AtomParser("kg2"))    == "Atom(1.000e+00 kg=2)"

def test_quantity():
    
    assert str(UnitSolver("kg*m2/s2"))         == "Atom(1.000e+00 kg=1 m=2 s=-2)"
    assert str(UnitSolver("12/4*kg/(m2*s2)"))  == "Atom(3.000e+00 kg=1 m=-2 s=-2)"



import sys, os
import pytest
import numpy as np
sys.path.insert(0, 'src')

from scinumtools.dip import DIP
from scinumtools.dip.settings import Format
from scinumtools.dip.datatypes import FloatType, IntegerType, StringType

def parse(code):
    with DIP() as p:
        p.from_string(code)
        return p.parse().data(verbose=True,format=Format.TYPE)

def test_import_nodes():
    data = parse('''
$source nodes = blocks/nodes.dip
    
{nodes}                        # base import
box
  {nodes}                      # import into a group node
basket.bag {nodes}             # import into a namespace
    ''')
    np.testing.assert_equal(data,{
        'fruits': IntegerType(0),
        'vegies': IntegerType(1),
        'vegies.potato': FloatType(200.0, 'g'),
        'box.fruits': IntegerType(0),
        'box.vegies': IntegerType(1),
        'box.vegies.potato': FloatType(200.0, 'g'),
        'basket.bag.fruits': IntegerType(0),
        'basket.bag.vegies': IntegerType(1),
        'basket.bag.vegies.potato': FloatType(200.0, 'g'),
    })

def test_import_query_remote():
    data = parse('''
$source nodes = blocks/nodes.dip

bag {nodes?*}                # import all
bowl 
  {nodes?fruits}             # selecting a specific node
  {nodes?vegies.potato}      # selecting a specific subnode
plate {nodes?vegies.*}       # selecting all subnodes
    ''')
    np.testing.assert_equal(data,{
        'bag.fruits': IntegerType(0),
        'bag.vegies': IntegerType(1),
        'bag.vegies.potato': FloatType(200.0, 'g'),
        'bowl.fruits': IntegerType(0),
        'bowl.potato': FloatType(200.0, 'g'),
        'plate.potato': FloatType(200.0, 'g'),
    })

def test_source_inject():
    data = parse('''
file str = blocks/nodes.dip
$source nodes = {?file}
{nodes?fruits}              
    ''')
    np.testing.assert_equal(data,{
        'file':   StringType('blocks/nodes.dip'),
        'fruits': IntegerType(0),
    })

def test_source_import():
    data = parse('''
$source file = blocks/nodes.dip
$source {file?blocks}
{blocks?energy}              
    ''')
    np.testing.assert_equal(data,{
        'energy': FloatType(13, 'J')
    })
    data = parse('''
$source file = blocks/nodes.dip
$source {file?*}
{blocks?energy}              
    ''')
    np.testing.assert_equal(data,{
        'energy': FloatType(13, 'J')
    })
    with pytest.raises(Exception) as e_info:
        parse('''
$source file = blocks/nodes.dip
$source {file?books}
        ''')
    assert e_info.value.args[0] == "Requested source does not exists:"
    with pytest.raises(Exception) as e_info:
        parse('''
$source blocks = blocks/nodes.dip
$source {blocks?blocks}
        ''')
    assert e_info.value.args[0] == "Reference source alread exists:"
    
def test_import_query_local():
    data = parse('''
icecream 
  waffle str = 'standard'
  scoops
    strawberry int = 1 #some comment
    chocolate int = 2

bowl
  {?icecream.scoops.*}      # select subnodes from current file
plate {?icecream.waffle}    # select specific node from current file
    ''')
    np.testing.assert_equal(data,{
        'icecream.waffle':            StringType('standard'),
        'icecream.scoops.strawberry': IntegerType(1),
        'icecream.scoops.chocolate':  IntegerType(2),
        'bowl.strawberry':            IntegerType(1),
        'bowl.chocolate':             IntegerType(2),
        'plate.waffle':               StringType('standard'),
    })

def test_injection_local():
    data = parse('''
size1 float = 34 cm       # standard definition
size2 float = {?size1} m  # definition using import with other units
size3 float = {?size2}    # definition using import with same units
size1 = {?size2}          # modifying by import
    ''')
    np.testing.assert_equal(data,{
        'size1': FloatType(3400, 'cm'),
        'size2': FloatType(34, 'm'),
        'size3': FloatType(34, 'm'),
    })

def test_injection_remote():
    data = parse('''
$source query = blocks/query.dip
    
energy float = 34 erg
energy float = {query?energy}  # import with a type
energy = {query?energy} eV     # import value only but set a different unit
energy = {query?energy}        # import both value and unit
    ''')
    np.testing.assert_equal(data,{
        'energy': FloatType(13.0e7, 'erg')     # converted to original units
    })
    with pytest.raises(Exception) as e_info:
        parse('''
        $source query = blocks/query.dip
        
        energy float = {query?*}
        ''')
    assert e_info.value.args[0] == "Path returned invalid number of nodes:"
    assert e_info.value.args[1] == "query?*"

def test_injection_element():
    data = parse('''
sizes float[3] = [34,23.34,1e34] cm      
mysize float[2] = {?sizes}[:2]                # slicing range of values
masses float[2,2] = [[34,23.34],[1,1e34]] cm    
mymass float[2] = {?masses}[:,1]              # slicing multiple range of values
    ''')
    np.testing.assert_equal(data,{
        'sizes': FloatType([3.400e+01, 2.334e+01, 1.000e+34], 'cm'),
        'mysize': FloatType([3.400e+01, 2.334e+01], 'cm'),
        'masses': FloatType([[34,23.34],[1,1e34]], 'cm'),
        'mymass': FloatType([23.34,1e34], 'cm')
    })
    with pytest.raises(Exception) as e_info:
        parse('''
sizes float[3] = [34,23.34,1e34] cm       # standard definition
mysize float = {?sizes}[:2]
        ''')
    assert e_info.value.args[0] == "Array value set to scalar node:"
    
def test_injection_array():
    data = parse('''
$source matrix = blocks/matrix.txt
$source query = blocks/query.dip
    
blocks                                  # block imports into a group node
  matrix1 int[3,4] = {matrix}           # import a text file
  matrix2 int[3,4] = {query?matrix}     # import value of a specific node
  matrix3 float[3,4] = {matrix}         # import a text file
  matrix4 float[3,4] = {query?matrix}   # import value of a specific node
    ''')
    np.testing.assert_equal(data,{
        'blocks.matrix1': IntegerType([[4234,   34,   35,   34],
                                        [ 234,   34,  644,   43],
                                        [ 353, 2356,  234,    3]]),
        'blocks.matrix2': IntegerType([[4234,   34,   35,   34],
                                       [ 234,   34,  644,   43],
                                       [ 353, 2356,  234,    3]]),
        'blocks.matrix3': FloatType([[4234,   34,   35,   34],
                                     [ 234,   34,  644,   43],
                                     [ 353, 2356,  234,    3]]),
        'blocks.matrix4': FloatType([[4234,   34,   35,   34],
                                     [ 234,   34,  644,   43],
                                     [ 353, 2356,  234,    3]]),
    })
    
def test_injection_table():
    data = parse('''
$source table = blocks/table.txt  
$source query = blocks/query.dip
    
blocks                             # block imports into a group node
  table1 table = {table}           # import a text file
  table2 table = {query?table}     # import value of a specific node
    ''')
    np.testing.assert_equal(data,{
        'blocks.table1.x': FloatType([0.234, 1.355, 2.535, 3.255, 4.455], 'm'),
        'blocks.table1.y': FloatType([0.234 , 1.43  , 2.423 , 3.2355, 4.2356], 'm'),
        'blocks.table2.x': FloatType([0.234, 1.355, 2.535, 3.255, 4.455], 'm'),
        'blocks.table2.y': FloatType([0.234 , 1.43  , 2.423 , 3.2355, 4.2356], 'm'),
    })
    
def test_injection_text():
    data = parse('''
$source text = blocks/text.txt
$source query = blocks/query.dip
    
blocks                         # block imports into a group node
  text1 str = {text}           # import a text file
  text2 str = {query?text}     # import value of a specific node
    ''')
    np.testing.assert_equal(data,{
        'blocks.text1': StringType('This is a block text\nwith multiple lines\nthat will be loaded to a\nstring node.\n'),
        'blocks.text2': StringType('This is a block text\nwith multiple lines\nthat will be loaded to a\nstring node.'),
    })

            
if __name__ == "__main__":
    # Specify wich test to run
    test = sys.argv[1] if len(sys.argv)>1 else True

    # Loop through all tests
    for fn in dir(sys.modules[__name__]):
        if fn[:5]=='test_' and (test is True or test==fn[5:]):
            print(f"\nTesting: {fn}\n")
            locals()[fn]()



import numpy as np
import os
import sys
sys.path.insert(0, 'src')

from scinumtools import RowCollector
from scinumtools.units import Dimensions
from scinumtools.units.settings import *
from scinumtools.units.unit_types import *

path_docs_static = 'docs/source/_static/tables'

# Create table of unit prefixes
rc = RowCollector(['Symbol','Name','Magnitude'])
for symbol, prefix in UNIT_PREFIXES.items():
    exponent = int(np.log10(prefix.magnitude))
    rc.append([symbol, prefix.name, ":math:`10^{"+f"{exponent:d}"+"}`"])
df = rc.to_dataframe()
df.to_csv(f"{path_docs_static}/prefixes.csv", index=False)

# create table of base units
rc = RowCollector(['Symbol','Name','Prefixes'])
for symbol, unit in UNIT_STANDARD.items():
    if unit.definition is not None:
        continue
    if unit.prefixes is True:
        prefixes = 'all'
    elif isinstance(unit.prefixes, list):
        prefixes = ", ".join([f"{p}{symbol}" for p in unit.prefixes])
    else:
        prefixes = ''
    rc.append([symbol, unit.name, prefixes])
df = rc.to_dataframe()
df.to_csv(f"{path_docs_static}/unit_base.csv", index=False)

# create table of standard units
rc = RowCollector(['Symbol','Name','Definition','Prefixes'])
for symbol, unit in UNIT_STANDARD.items():
    if not isinstance(unit.definition, str):
        continue
    if symbol in ['K','degR','AR','PR']:
        continue
    if symbol.startswith('['):
        continue
    if unit.prefixes is True:
        prefixes = 'all'
    elif isinstance(unit.prefixes, list):
        prefixes = ", ".join([f"{p}{symbol}" for p in unit.prefixes])
    else:
        prefixes = ''
    if len(rc) and unit.name in rc.Name:
        idx = rc.Name.index(unit.name)
        rc.Symbol[idx] += ", "+symbol
        rc.Prefixes[idx] += ", "+prefixes if rc.Prefixes[idx] else ""
    else:
        rc.append([symbol, unit.name, unit.definition, prefixes])
df = rc.to_dataframe()
df.to_csv(f"{path_docs_static}/unit_standard.csv", index=False)

# create table of constants
rc = RowCollector(['Symbol','Name','Definition'])
for symbol, unit in UNIT_STANDARD.items():
    if not symbol.startswith('['):
        continue
    if len(rc) and unit.name in rc.Name:
        idx = rc.Name.index(unit.name)
        rc.Symbol[idx] += ", "+symbol
    else:
        rc.append([symbol, unit.name, unit.definition])
df = rc.to_dataframe()
df.to_csv(f"{path_docs_static}/constants.csv", index=False)

# create table of logarithmic units
rc = RowCollector(['Symbol','Name','Prefixes'])
for symbol, unit in UNIT_STANDARD.items():
    if symbol not in ['AR','PR'] and unit.definition != LogarithmicUnitType:
        continue
    if unit.prefixes is True:
        prefixes = 'all'
    elif isinstance(unit.prefixes, list):
        prefixes = ", ".join([f"{p}{symbol}" for p in unit.prefixes])
    else:
        prefixes = ''
    if len(rc) and unit.name in rc.Name:
        idx = rc.Name.index(unit.name)
        rc.Symbol[idx] += ", "+symbol
        rc.Prefixes[idx] += ", "+prefixes if rc.Prefixes[idx] else ""
    else:
        rc.append([symbol, unit.name, prefixes])
df = rc.to_dataframe()
df.to_csv(f"{path_docs_static}/unit_logarithmic.csv", index=False)

# create table of temperature units
rc = RowCollector(['Symbol','Name','Prefixes'])
for symbol, unit in UNIT_STANDARD.items():
    if symbol not in ['K','degR'] and unit.definition != TemperatureUnitType:
        continue
    if unit.prefixes is True:
        prefixes = 'all'
    elif isinstance(unit.prefixes, list):
        prefixes = ", ".join([f"{p}{symbol}" for p in unit.prefixes])
    else:
        prefixes = ''
    rc.append([symbol, unit.name, prefixes])
df = rc.to_dataframe()
df.to_csv(f"{path_docs_static}/unit_temperature.csv", index=False)

# create table of physical quantities SI
rc = RowCollector(['Symbol','Name','Definition','Dimensions'])
for q in range(len(QUANTITY_LIST)):
    name = QUANTITY_LIST.name[q]
    symbol = QUANTITY_LIST.symbol[q]
    definition = QUANTITY_LIST.SI[q]
    magnitude, dimensions = QUANTITY_UNITS[f"#S{symbol}"]
    rc.append([f"#S{symbol}", f"SI.{name}", definition, dimensions])
df = rc.to_dataframe()
df.to_csv(f"{path_docs_static}/quantities_si.csv", index=False)

# create table of physical quantities CGS
rc = RowCollector(['Symbol','Name','Definition'])
for q in range(len(QUANTITY_LIST)):
    symbol = QUANTITY_LIST.symbol[q]
    if f"#C{symbol}" not in QUANTITY_UNITS:
        continue
    name = QUANTITY_LIST.name[q]
    definition = QUANTITY_LIST.CGS[q]
    rc.append([f"#C{symbol}", f"CGS.{name}", definition])
df = rc.to_dataframe()
df.to_csv(f"{path_docs_static}/quantities_cgs.csv", index=False)

# create table of physical quantities AU
rc = RowCollector(['Symbol','Name','Definition'])
for q in range(len(QUANTITY_LIST)):
    symbol = QUANTITY_LIST.symbol[q]
    if f"#A{symbol}" not in QUANTITY_UNITS:
        continue
    name = QUANTITY_LIST.name[q]
    definition = QUANTITY_LIST.AU[q]
    rc.append([f"#A{symbol}", f"AU.{name}", definition])
df = rc.to_dataframe()
df.to_csv(f"{path_docs_static}/quantities_au.csv", index=False)

